/*
FreeRTOS+TCP V2.0.1
Copyright (C) 2017 Amazon.com, Inc. or its affiliates.  All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 http://aws.amazon.com/freertos
 http://www.FreeRTOS.org
*/

/* Standard includes. */
#include <stdio.h>
#include <stdint.h>

/* FreeRTOS includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

/* FreeRTOS+TCP includes. */
#include "FreeRTOS_IP.h"
#include "FreeRTOS_UDP_IP.h"
#include "FreeRTOS_Sockets.h"
#include "NetworkBufferManagement.h"
#include "FreeRTOS_IP_Private.h"

/* Hardware includes. */
#include "r_ether_rx_if.h"

/* Demo includes. */
#include "NetworkInterface.h"

/* include FIT module */
#include "platform.h"
#include "r_cmt_rx_if.h"

#if ipconfigETHERNET_DRIVER_FILTERS_FRAME_TYPES != 1
	#define ipCONSIDER_FRAME_FOR_PROCESSING( pucEthernetBuffer ) eProcessBuffer
#else
	#define ipCONSIDER_FRAME_FOR_PROCESSING( pucEthernetBuffer ) eConsiderFrameForProcessing( ( pucEthernetBuffer ) )
#endif

/* When a packet is ready to be sent, if it cannot be sent immediately then the
task performing the transmit will block for niTX_BUFFER_FREE_WAIT
milliseconds.  It will do this a maximum of niMAX_TX_ATTEMPTS before giving
up. */
#define niTX_BUFFER_FREE_WAIT	( ( TickType_t ) 2UL / portTICK_PERIOD_MS )
#define niMAX_TX_ATTEMPTS		( 5 )

/* The length of the queue used to send interrupt status words from the
interrupt handler to the deferred handler task. */
#define niINTERRUPT_QUEUE_LENGTH	( 10 )

/*-----------------------------------------------------------*/

/*
 * A deferred interrupt handler task that processes
 */
extern void vEMACHandlerTask( void *pvParameters );

/*-----------------------------------------------------------*/

/* The queue used to communicate Ethernet events with the IP task. */
extern QueueHandle_t xNetworkEventQueue;

/* The semaphore used to wake the deferred interrupt handler task when an Rx
interrupt is received. */
SemaphoreHandle_t xEMACRxEventSemaphore = NULL;
/*-----------------------------------------------------------*/

void check_ether_link(void);
void receive_check(void);

static TimerHandle_t timer_handle;
static uint32_t timer_id;
static TaskHandle_t ether_receive_check_task_handle;

BaseType_t xNetworkInterfaceInitialise( void )
{
	BaseType_t return_code = pdFALSE;
	uint32_t channel;

	if(lan_open() == 0)
	{
		return_code = pdTRUE;
	}

	while(ETHER_SUCCESS != R_ETHER_CheckLink_ZC(0));
	R_BSP_SoftwareDelay(3, BSP_DELAY_SECS);

	return_code = xTaskCreate(receive_check, "ETHER_RECEIVE_CHECK_TASK", 100/* stack size (word) */, 0,  configMAX_PRIORITIES, &ether_receive_check_task_handle);
	if(return_code == xTaskCreate)
	{
        /* The task was created.  Use the task's handle to delete the task. */
        vTaskDelete( ether_receive_check_task_handle );
	}

	timer_id = 1;
	timer_handle = xTimerCreate("ETHER_TIMER", 10, pdTRUE, &timer_id, check_ether_link);
	xTimerStart(timer_handle, 0);

	return return_code;
}
/*-----------------------------------------------------------*/

BaseType_t xNetworkInterfaceOutput( NetworkBufferDescriptor_t * const pxNetworkBuffer )
{
	int16_t return_code;

	/* 簡易実装しました。あとでゼロコピー対応に直します。（シェルティ） */
	static uint8_t tmp_buf[1514];
	memcpy(tmp_buf, pxNetworkBuffer->pucEthernetBuffer, pxNetworkBuffer->xDataLength);
	return_code = lan_write(0, tmp_buf, 14, tmp_buf + 14, pxNetworkBuffer->xDataLength - 14);
	if(return_code > 0)
	{
		/* Call the standard trace macro to log the send event. */
		iptraceNETWORK_INTERFACE_TRANSMIT();

		vReleaseNetworkBufferAndDescriptor(pxNetworkBuffer);
	}
	else
	{
		/* なぜか lan_write()が-5を返す。あとで確認する。★ */
		vReleaseNetworkBufferAndDescriptor(pxNetworkBuffer);
	}
	return 0;
}
/*-----------------------------------------------------------*/

uint32_t ulRand(void)
{
	/* 後でハードウェア実装(暗号器使用)に変更します (シェルティ) */
	uint32_t tmp;
	get_random_number(tmp, 4);
	return tmp;
}

void vNetworkInterfaceAllocateRAMToBuffers( NetworkBufferDescriptor_t pxNetworkBuffers[ ipconfigNUM_NETWORK_BUFFER_DESCRIPTORS ] )
{
	uint32_t ul;
    uint8_t *buffer_address;
    buffer_address = __sectop("B_ETHERNET_BUFFERS_1");

    /* ------------------------------------------------------------------
     * RX MCU Ether driver buffer specification
     * ------------------------------------------------------------------
     *
     * top -> B_ETHERNET_BUFFERS_1 |receive buffer 1| (each buffers are always aligned at 32bit)
     *                             |receive buffer 2|
     *                             |       :        |  ETHER_CFG_EMAC_RX_DESCRIPTORS
     *                             |transmit buffer1|
     *                             |transmit buffer2|
     *                             |       :        |  ETHER_CFG_EMAC_TX_DESCRIPTORS
     * */

	for( ul = 0; ul < ipconfigNUM_NETWORK_BUFFER_DESCRIPTORS; ul++ )
	{
		pxNetworkBuffers[ul].pucEthernetBuffer = (buffer_address + (ETHER_CFG_BUFSIZE * ul));
	}
}

void check_ether_link(void)
{
	R_ETHER_LinkProcess(0);
}

void receive_check(void)
{
	bool pktSuccess, pktLost;
	IPStackEvent_t xRxEvent;
	uint8_t *buffer_pointer;
	int32_t received_data_size;
	NetworkBufferDescriptor_t *pxBufferDescriptor;

	while(1)
	{
		vTaskDelay(10);
		received_data_size = lan_read(0, &buffer_pointer);

		if(received_data_size > 0)
		{
			pxBufferDescriptor = pxGetNetworkBufferWithDescriptor( received_data_size, 0 );
			if(pxBufferDescriptor != NULL)
			{
				memcpy(pxBufferDescriptor->pucEthernetBuffer, buffer_pointer, received_data_size);
				pxBufferDescriptor->xDataLength = received_data_size;
				xRxEvent.eEventType = eNetworkRxEvent;
				xRxEvent.pvData = pxBufferDescriptor;
				// Send the data to the TCP/IP stack
				if( xSendEventStructToIPTask( &xRxEvent, 0 ) == pdFALSE )
				{   // failed
					vReleaseNetworkBufferAndDescriptor( pxBufferDescriptor );
					pktLost = true;
				}
				else
				{   // success
					pktSuccess = true;
				}
			}
		}
		rcv_buff_release(0, buffer_pointer);
	}
}

